#
# Copyright (c) 2025 CESNET z.s.p.o.
#
# This file is a part of oarepo-model (see http://github.com/oarepo/oarepo-model).
#
# oarepo-model is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
#
"""Module providing preset for registering resolvers."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, override

from invenio_access.permissions import system_identity
from invenio_drafts_resources.services import RecordService as RecordServiceWithDrafts
from invenio_pidstore.errors import PIDDoesNotExistError
from invenio_rdm_records.requests.entity_resolvers import (
    RDMRecordServiceResultProxy,
    RDMRecordServiceResultResolver,
)
from invenio_records_resources.references.entity_resolvers.results import (
    ServiceResultResolver as InvenioServiceResultResolver,
)
from oarepo_model.customizations import (
    AddEntryPoint,
    AddModule,
    AddToModule,
    Customization,
)
from oarepo_model.presets import Preset
from oarepo_runtime.typing import record_from_result
from sqlalchemy.exc import NoResultFound

if TYPE_CHECKING:
    from collections.abc import Generator

    from invenio_drafts_resources.records import Draft
    from invenio_records_resources.records import Record
    from invenio_records_resources.references import RecordResolver
    from invenio_records_resources.services.records.results import RecordItem
    from oarepo_model.builder import InvenioModelBuilder
    from oarepo_model.model import InvenioModel


class RecordServiceResultProxy(RDMRecordServiceResultProxy):
    """Proxy to resolve a record service result."""

    @override
    def _resolve(self) -> dict[str, Any]:
        """Resolve the result item from the proxy's reference dict."""
        pid_value = self._parse_ref_dict_id()
        if isinstance(self.service, RecordServiceWithDrafts):
            try:
                draft = self.service.read_draft(system_identity, pid_value)
            except (PIDDoesNotExistError, NoResultFound):
                record = self._get_record(pid_value)
            else:
                record = (
                    draft if not record_from_result(draft).is_published else self._get_record(pid_value)  # type: ignore[reportAttributeAccessIssue]
                )
        else:
            record = self._get_record(pid_value)

        return record.to_dict()  # type: ignore[no-any-return]


class RecordServiceResultResolver(RDMRecordServiceResultResolver):
    """Service result resolver for draft records."""

    def __init__(
        self,
        service_id: str,
        type_key: str,
        proxy_cls: type[RecordServiceResultProxy] = RecordServiceResultProxy,
        item_cls: type[RecordItem] | None = None,
        record_cls: type[Record] | None = None,
    ):
        """Initialize the resolver."""
        super(RDMRecordServiceResultResolver, self).__init__(service_id, type_key, proxy_cls, item_cls, record_cls)

    @property
    @override
    def draft_cls(self) -> type[Draft] | None:
        """Get specified draft class or from service."""
        service = self.get_service()
        return service.draft_cls if isinstance(service, RecordServiceWithDrafts) else None

    @override
    def matches_entity(self, entity: Any) -> bool:
        """Check if the entity is a draft."""
        if self.draft_cls and isinstance(entity, self.draft_cls):
            return True

        return InvenioServiceResultResolver.matches_entity(self, entity=entity)  # type: ignore[no-any-return]


class RegisterResolversPreset(Preset):
    """Preset for registering resolvers."""

    @override
    def apply(
        self,
        builder: InvenioModelBuilder,
        model: InvenioModel,
        dependencies: dict[str, Any],
    ) -> Generator[Customization]:
        def register_entity_resolver() -> RecordResolver:
            service_id = builder.model.base_name
            runtime_dependencies = builder.get_runtime_dependencies()
            resolver = runtime_dependencies.get("RecordResolver")
            return resolver(
                record_cls=runtime_dependencies.get("Record"),
                service_id=service_id,
                type_key=service_id,
                proxy_cls=runtime_dependencies.get("RecordProxy"),
            )

        def register_notification_resolver() -> RecordServiceResultResolver:
            service_id = builder.model.base_name
            return RecordServiceResultResolver(
                service_id=service_id,
                type_key=service_id,
                proxy_cls=RecordServiceResultProxy,
            )

        # just invenio things
        register_notification_resolver.type_key = builder.model.base_name  # type: ignore[attr-defined]

        yield AddModule("resolvers", exists_ok=True)
        yield AddToModule(
            "resolvers",
            "register_entity_resolver",
            staticmethod(register_entity_resolver),
        )
        yield AddToModule(
            "resolvers",
            "register_notification_resolver",
            staticmethod(register_notification_resolver),
        )
        yield AddEntryPoint(
            group="invenio_requests.entity_resolvers",
            name=f"{model.base_name}_requests",
            value="resolvers:register_entity_resolver",
            separator=".",
        )
        yield AddEntryPoint(
            group="invenio_notifications.entity_resolvers",
            name=f"{model.base_name}_requests",
            value="resolvers:register_notification_resolver",
            separator=".",
        )



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""OARepo extension to invenio-requests service."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, override

from invenio_drafts_resources.records.api import Record
from invenio_i18n import _
from invenio_records_resources.services import LinksTemplate
from invenio_records_resources.services.base.links import EndpointLink
from invenio_records_resources.services.uow import IndexRefreshOp, unit_of_work
from invenio_requests import current_request_type_registry
from invenio_requests.services import RequestsService

from oarepo_requests.errors import CustomHTTPJSONException, UnknownRequestTypeError
from oarepo_requests.proxies import current_oarepo_requests
from oarepo_requests.services.results import (
    RequestTypesList,
    StringEntityResolverExpandableField,
)
from oarepo_requests.utils import (
    allowed_request_types_for_record,
    resolve_reference_dict,
)

if TYPE_CHECKING:
    from datetime import datetime

    from flask_principal import Identity
    from invenio_db.uow import UnitOfWork
    from invenio_requests.services.requests.results import RequestItem
    from invenio_requests.services.results import (
        EntityResolverExpandableField,
        MultiEntityResolverExpandableField,
    )


class OARepoRequestsService(RequestsService):
    """OARepo extension to invenio-requests service."""

    @property
    def expandable_fields(
        self,
    ) -> list[EntityResolverExpandableField | MultiEntityResolverExpandableField]:
        """Get expandable fields."""
        return [
            *super().expandable_fields,
            StringEntityResolverExpandableField("payload.created_topic"),
        ]

    @unit_of_work()
    @override
    def create(
        self,
        identity: Identity,
        data: dict[str, Any] | None,
        request_type: str,
        receiver: dict[str, str] | Any | None = None,
        *args: Any,
        creator: dict[str, str] | Any | None = None,
        topic: Record | None = None,
        expires_at: datetime | None = None,
        uow: UnitOfWork,
        expand: bool = False,
        **kwargs: Any,
    ) -> RequestItem:
        """Create a request.

        :param identity: Identity of the user creating the request.
        :param data: Data of the request.
        :param request_type: Type of the request.
        :param receiver: Receiver of the request. If unfilled, a default receiver from workflow is used.
        :param creator: Creator of the request.
        :param topic: Topic of the request.
        :param expires_at: Expiration date of the request.
        :param uow: Unit of work.
        :param expand: Expand the response.
        :param args: Additional arguments.
        :param kwargs: Additional keyword arguments.
        """
        # TODO: invenio suggest None topic can be here but we do not expect it
        if topic is None:
            raise ValueError("")
        type_ = (
            current_request_type_registry.lookup(request_type, quiet=True)
            if isinstance(request_type, str)
            else request_type
        )
        if not type_:
            raise UnknownRequestTypeError(request_type)
        data = data or {}
        if receiver is None:
            # if explicit creator is not passed, use current identity - this is in sync with invenio_requests
            receiver = current_oarepo_requests.default_request_receiver(
                identity, type_, topic, creator or identity, data
            )
        if "payload" not in data and type_.payload_schema:
            data["payload"] = {}
        schema = self._wrap_schema(type_.marshmallow_schema())
        data, errors = schema.load(
            data,
            context={"identity": identity},
            raise_errors=False,
        )
        if errors:
            raise CustomHTTPJSONException(
                description=_("Action could not be performed due to validation request fields validation errors."),
                request_payload_errors=errors,
                code=400,
            )
        if hasattr(type_, "can_create"):
            # raise exception if can't
            type_.can_create(identity, data, receiver, topic, creator)
        # TODO: typing does not allow receiver to be None even though invenio code suggests it
        result = super().create(
            identity=identity,
            data=data,
            request_type=type_,
            receiver=receiver,  # type: ignore[reportArgumentType]
            creator=creator,
            topic=topic,
            expand=expand,
            uow=uow,
        )
        uow.register(IndexRefreshOp(indexer=self.indexer, index=self.record_cls.index))  # type: ignore[reportArgumentType]
        return result

    def applicable_request_types(self, identity: Identity, topic: Record | dict[str, str]) -> RequestTypesList:
        """Get applicable request types for a record."""
        topic = resolve_reference_dict(topic) if not isinstance(topic, Record) else topic
        if not isinstance(topic, Record):
            raise TypeError("Trying to find applicable request types on non-record entity")

        allowed_request_types = allowed_request_types_for_record(identity, topic)
        return RequestTypesList(
            service=self,
            identity=identity,
            results=list(allowed_request_types.values()),
            links_tpl=LinksTemplate({"self": EndpointLink("requests.applicable_request_types")}),
            record=topic,
        )



#
# Copyright (c) 2025 CESNET z.s.p.o.
#
# This file is a part of oarepo-requests (see https://github.com/oarepo/oarepo-requests).
#
# oarepo-requests is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
#

"""Components for request type and request expansion in record responses.

This module provides service components that handle the expansion of request-related
information when records are retrieved. The components are used to:

1. Expand request types available for a record (RequestTypesComponent)
2. Expand existing requests associated with a record (RequestsComponent)

Both components work with the 'expand' parameter in record retrieval operations
and add their data under the 'expanded' section of the response.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from invenio_records_resources.services.errors import PermissionDeniedError
from oarepo_runtime.services.results import ResultComponent

from oarepo_requests.services.results import serialize_request_types
from oarepo_requests.utils import allowed_request_types_for_record, search_requests

if TYPE_CHECKING:
    from flask_principal import Identity
    from invenio_records_resources.records import Record


class RequestTypesComponent(ResultComponent):
    """Component for expanding request types."""

    def update_data(self, identity: Identity, record: Record, projection: dict, expand: bool) -> None:
        """Expand request types if requested."""
        if not expand:
            return
        allowed_request_types = allowed_request_types_for_record(identity, record)
        request_types_list = serialize_request_types(allowed_request_types, identity, record)
        projection["expanded"]["request_types"] = request_types_list


class RequestsComponent(ResultComponent):
    """Component for expanding requests on a record."""

    def update_data(self, identity: Identity, record: Record, projection: dict, expand: bool) -> None:
        """Expand requests if requested."""
        if not expand:
            return
        try:
            requests = list(search_requests(identity, record))
        except PermissionDeniedError:
            requests = []
        projection["expanded"]["requests"] = requests



#
# Copyright (C) 2025 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""OARepo requests search."""

from __future__ import annotations

from invenio_requests.resources.requests.config import RequestSearchRequestArgsSchema
from invenio_requests.services.requests.config import RequestSearchOptions
from marshmallow import fields

from oarepo_requests.services.facets.params import (
    IsClosedParam,
    RequestAllAvailableFilterParam,
    RequestNotOwnerFilterParam,
    RequestOwnerFilterParam,
)


class EnhancedRequestSearchOptions(RequestSearchOptions):
    """Searched options enhanced with additional filters."""

    # TODO: params_interpreters_cls: tuple[type[ParamInterpreter], ...]
    #  in stubs is not assignable to partial[ParamInterpreter]
    params_interpreters_cls = (  # type: ignore[reportAssignmentType]
        *RequestSearchOptions.params_interpreters_cls,
        RequestOwnerFilterParam.factory("mine", "created_by.user"),
        RequestNotOwnerFilterParam.factory("assigned", "created_by.user"),
        RequestAllAvailableFilterParam.factory("all"),
        IsClosedParam.factory("is_closed"),
    )


class ExtendedRequestSearchRequestArgsSchema(RequestSearchRequestArgsSchema):
    """Marshmallow schema for the extra filters."""

    mine = fields.Boolean()
    assigned = fields.Boolean()
    all = fields.Boolean()
    is_closed = fields.Boolean()



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Enhancements to the request schema."""

from __future__ import annotations

from contextvars import ContextVar
from typing import Any

import marshmallow as ma
from invenio_drafts_resources.records.api import Record
from invenio_records_resources.services.base.links import (
    EndpointLink,
    LinksTemplate,
)
from invenio_requests.services.schemas import GenericRequestSchema
from marshmallow import fields

from oarepo_requests.utils import ref_to_str, reference_entity

request_type_identity_ctx: ContextVar[Any] = ContextVar("oarepo_requests.request_type_identity", default=None)
request_type_record_ctx: ContextVar[Any] = ContextVar("oarepo_requests.request_type_record", default=None)


def get_links_schema() -> ma.fields.Dict:
    """Get links schema."""
    return ma.fields.Dict(keys=ma.fields.String())  # value is either string or dict of strings (for actions)


class RequestTypeSchema(ma.Schema):
    """Request type schema."""

    type_id = ma.fields.String()
    """Type ID of the request type."""

    links = get_links_schema()
    """Links to the request type."""

    @ma.post_dump
    def _create_link(self, data: dict, **kwargs: Any) -> dict:  # noqa ARG002
        if "links" in data:
            return data
        type_id = data["type_id"]
        identity = request_type_identity_ctx.get()
        record = request_type_record_ctx.get()
        topic_ref = ref_to_str(reference_entity(record) if isinstance(record, Record) else record)
        link = EndpointLink("requests.create_via_url", params=["topic", "request_type"])
        template = LinksTemplate({"create": link}, context={"topic": topic_ref, "request_type": type_id})
        data["links"] = {"actions": template.expand(identity, record)}
        return data


class NoReceiverAllowedGenericRequestSchema(GenericRequestSchema):
    """A mixin that allows serialization of requests without a receiver."""

    receiver = fields.Dict(allow_none=True)


class RequestsSchemaMixin:
    """A mixin that allows serialization of requests together with their request type."""

    requests = ma.fields.List(ma.fields.Nested(NoReceiverAllowedGenericRequestSchema))
    request_types = ma.fields.List(ma.fields.Nested(RequestTypeSchema))



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Helper functions to convert identity to entity references."""

from __future__ import annotations

from typing import TYPE_CHECKING

# TODO: possibly not used
if TYPE_CHECKING:
    from flask_principal import Identity


def user_mappings(identity: Identity) -> list[dict[str, str]]:
    """Convert identity to entity references of type "user"."""
    return [{"user": identity.id}]


def group_mappings(identity: Identity) -> list[dict[str, str]]:
    """Convert groups of the identity to entity references of type "group"."""
    roles = [n.value for n in identity.provides if n.method == "role"]
    return [{"group": role_id} for role_id in roles]
